//# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
//# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
//# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
//# SPDX-License-Identifier: MIT
#pragma once

#include "columns.hpp"
#include "data_ptr.hpp"
#include "equality_comparable.hpp"
#include "iterators.hpp"
#include "less_than_comparable.hpp"
#include "resizable.hpp"
#include "rows.hpp"
#include "size_and_capacity.hpp"
#include "spans.hpp"
#include "swappable.hpp"
